use crate::listener::*;
use crate::*;
use std::collections::{HashMap, HashSet};
use std::env;
use std::fmt;
use std::io::prelude::*;
use std::io::{BufRead, BufReader, Error, ErrorKind, Result};
use std::net::SocketAddr;
use std::os::unix::net::UnixStream;
use std::path::PathBuf;
use std::{thread, time};

pub(crate) fn gen_events(
    state: &HashMap<String, Peer>,
    prev: &HashMap<String, Peer>,
    listeners: &Vec<Box<EventListener>>,
    timeout: time::Duration,
) {
    let side_by_side = {
        state
            .keys()
            .map(String::as_ref)
            .chain(prev.keys().map(String::as_ref))
            .collect::<HashSet<&str>>()
            .iter()
            .map(|p| (p.to_owned(), (prev.get(*p), state.get(*p))))
            .collect::<HashMap<&str, (Option<&Peer>, Option<&Peer>)>>()
    };
    for (_id, (prev, cur)) in side_by_side {
        /*if id != "HhRgEL2xsnEIqThSTUKLGaTXusorM1MFdjSSYvzBynY=" {
            continue;
        }
        println!("{} p {} c {}", _id, prev.is_some(), cur.is_some());*/
        match (prev, cur) {
            (Some(prev), Some(cur)) if prev.endpoint != cur.endpoint => {
                if let (Some(prev_addr), Some(_)) = (prev.endpoint, cur.endpoint) {
                    listeners.roaming(&cur, prev_addr);
                }
            }
            (Some(prev), Some(cur)) => {
                //shake > timeout && prev.shake < timeout -> listeners.iter().for_each(|l| l.disconnected(&cur));
                //shake < timeout && (prev.shake is none || prev.shake > timeout) -> listeners.iter().for_each(|l| l.connected(&cur));
                if let (Some(shake), Some(pshake)) = (cur.last_handshake, prev.last_handshake) {
                    if shake > timeout && pshake < timeout {
                        listeners.disconnected(&cur);
                    }
                    continue;
                }
                if let Some(shake) = cur.last_handshake {
                    if shake > timeout
                        && prev
                            .last_handshake
                            .map(|shake| shake > timeout)
                            .unwrap_or(true)
                    {
                        listeners.connected(&cur);
                    }
                    continue;
                }
            }
            #[cfg(addrem)]
            (None, Some(cur)) => listeners.added(&cur),
            #[cfg(addrem)]
            (Some(prev), None) => listeners.removed(&prev),
            #[cfg(not(addrem))]
            (None, Some(_cur)) => (),
            #[cfg(not(addrem))]
            (Some(_prev), None) => (),
            fail => {
                println!("{:?}", fail);
                unreachable!()
            }
        }
    }
}


#[cfg(test)]
mod test {
    use super::*;
    use crate::listener::*;
    use crate::*;
    use std::cell::RefCell;
    use std::collections::{HashMap, HashSet};
    use std::env;
    use std::fmt;
    use std::io::prelude::*;
    use std::io::{BufRead, BufReader, Error, ErrorKind, Result};
    use std::net::SocketAddr;
    use std::os::unix::net::UnixStream;
    use std::path::PathBuf;
    use std::{thread, time};

    struct TestListener {
        calls: RefCell<Vec<String>>,
    }
    impl TestListener {
        fn new() -> TestListener {
            Self::from(RefCell::new(vec![]))
        }

        fn from(calls: RefCell<Vec<String>>) -> TestListener {
            TestListener { calls: calls }
        }
    }
    impl EventListener for TestListener {
        fn added<'a>(&self, peer: &'a Peer) {
            self.calls
                .borrow_mut()
                .push(format!("add {}", peer.public_key));
        }

        fn connected<'a>(&self, peer: &'a Peer) {
            self.calls
                .borrow_mut()
                .push(format!("con {}", peer.public_key));
        }

        fn disconnected<'a>(&self, peer: &'a Peer) {
            self.calls
                .borrow_mut()
                .push(format!("dis {}", peer.public_key));
        }

        fn removed<'a>(&self, peer: &'a Peer) {
            self.calls
                .borrow_mut()
                .push(format!("rem {}", peer.public_key));
        }

        fn roaming<'a>(&self, peer: &'a Peer, previous_addr: SocketAddr) {
            self.calls
                .borrow_mut()
                .push(format!("rom {}", peer.public_key));
        }
    }

    fn listeners() -> (Vec<Box<EventListener>>, RefCell<Vec<String>>) {
        let calls: RefCell<Vec<String>> = RefCell::new(vec![]);
        (
            vec![
                Box::new(TestListener::from(calls.clone())),
                Box::new(LogListener),
            ],
            calls.clone(),
        )
    }

    fn b2h(b: &str) -> String {
        hex::encode(base64::decode(b).unwrap())
    }

    #[test]
    fn connected() {
        let bkey = "HhRgEL2xsnEIqThSTUKLGaTXusorM1MFdjSSYvzBynY=";
        let key = b2h(bkey);
        let prev: HashMap<String, Peer> = HashMap::new();
        let mut cur: HashMap<String, Peer> = HashMap::new();
        cur.insert(
            key.clone(),
            Peer::from_kv(&vec![
                ("public_key".to_string(), key.clone()),
                (
                    "last_handshake_time_nsec".to_string(),
                    (1000 * 1000 * 1).to_string(),
                ),
                ("endpoint".to_string(), "1.1.1.1:22222".to_string()),
            ])
            .unwrap(),
        );
        let (listeners, calls) = listeners();
        gen_events(&cur, &prev, &listeners, time::Duration::from_secs(3));
        assert_eq!(vec![["con", bkey].join(" ")], calls.borrow().clone());
    }

}
