use crate::Peer;
use std::net::SocketAddr;
use std::path::PathBuf;
use std::process::Command;
use std::thread;

pub trait EventListener {
    fn added<'a>(&self, peer: &'a Peer) {
        self.connected(peer);
    }

    fn connected<'a>(&self, peer: &'a Peer);

    fn disconnected<'a>(&self, peer: &'a Peer);

    fn removed<'a>(&self, peer: &'a Peer) {
        self.disconnected(peer)
    }

    fn roaming<'a>(&self, peer: &'a Peer, previous_addr: SocketAddr);
}

impl EventListener for Vec<Box<EventListener>> {
    fn added<'a>(&self, peer: &'a Peer) {
        self.iter().for_each(|l| l.added(&peer));
    }

    fn connected<'a>(&self, peer: &'a Peer) {
        self.iter().for_each(|l| l.connected(&peer));
    }

    fn disconnected<'a>(&self, peer: &'a Peer) {
        self.iter().for_each(|l| l.disconnected(&peer));
    }

    fn removed<'a>(&self, peer: &'a Peer) {
        self.iter().for_each(|l| l.removed(&peer));
    }

    fn roaming<'a>(&self, peer: &'a Peer, previous_addr: SocketAddr) {
        self.iter().for_each(|l| l.roaming(&peer, previous_addr));
    }
}

pub struct LogListener;

impl EventListener for LogListener {
    fn connected<'a>(&self, peer: &'a Peer) {
        println!("{} connected!", peer.public_key);
    }

    fn disconnected<'a>(&self, peer: &'a Peer) {
        println!("{} disconnected!", peer.public_key);
    }

    fn added<'a>(&self, peer: &'a Peer) {
        println!("{} added!", peer.public_key);
    }

    fn removed<'a>(&self, peer: &'a Peer) {
        println!("{} removed!", peer.public_key);
    }

    fn roaming<'a>(&self, peer: &'a Peer, previous_addr: SocketAddr) {
        println!(
            "{} roamed {} -> {}!",
            peer.public_key,
            previous_addr,
            peer.endpoint.unwrap()
        );
    }
}

pub struct ScriptListener {
    pub script: PathBuf,
}

impl ScriptListener {
    pub fn new(script: PathBuf) -> ScriptListener {
        ScriptListener { script }
    }

    fn mkcmd<'a>(&self, args: Vec<&'a str>) -> Command {
        let mut cmd = Command::new("/bin/sh");
        cmd.arg("-c");
        cmd.arg(format!("\"{}\"", args.join(" ")));
        cmd
    }

    fn call_sub<'a>(&self, args: Vec<&'a str>) {
        let mut cmd = self.mkcmd(args);
        thread::spawn(move || {
            cmd.spawn().expect("Failed to call Script hooḱ!");
        });
    }
}

impl EventListener for ScriptListener {
    fn connected<'a>(&self, peer: &'a Peer) {
        self.call_sub(vec!["connected", &peer.public_key]);
    }

    fn disconnected<'a>(&self, peer: &'a Peer) {
        self.call_sub(vec!["disconnected", &peer.public_key]);
    }

    fn added<'a>(&self, peer: &'a Peer) {
        self.call_sub(vec!["added", &peer.public_key]);
    }

    fn removed<'a>(&self, peer: &'a Peer) {
        self.call_sub(vec!["removed", &peer.public_key]);
    }

    fn roaming<'a>(&self, peer: &'a Peer, previous_addr: SocketAddr) {
        self.call_sub(vec![
            "roaming",
            &peer.public_key,
            &previous_addr.to_string(),
            &peer.endpoint.unwrap().to_string(),
        ]);
    }
}
